import java.util.Random;

class PacketDeparture implements Event {
    BasicLink link;

    public PacketDeparture(BasicLink l) {
	link = l;
    }

    public void process() {
	link.process_departure();
    }
}

class PacketDelivery implements Event {
    BasicLink link;
    Packet packet;

    public PacketDelivery(BasicLink l, Packet p) {
	link = l;
	packet = p;
    }

    public void process() {
	link.process_delivery(packet);
    }
}

class PacketQueue {
    private Packet [] Q;
    private int head;
    private int tail;

    PacketQueue(int size) {
	Q = new Packet[size];
	head = 0;
	tail = 0;
    }

    private int next(int i) {
	i += 1;
	if (i == Q.length)
	    i = 0;
	return i;
    }

    public boolean empty() {
	return head == tail;
    }

    public boolean full() {
	return next(tail) == head;
    }

    public Packet dequeue() {
	if (head == tail)
	    return null;
	Packet p = Q[head];
	head = next(head);
	return p;
    }

    public Packet first() {
	if (head == tail)
	    return null;
	return Q[head];
    }

    public void enqueue(Packet p) {
	int new_tail = next(tail);
	if (new_tail !=  head) {
	    Q[tail] = p;
	    tail = new_tail;
	} else {
	    System.out.printf("%f ", Simulation.time());
	    System.out.println("BasicLink: tx queue is full, packet dropped: "
			       + p.toString());
	}
    }
}

public class BasicLink implements Link {
    private double rate = 1000000; 	// 1Mbytes/second
    private double delay = 0.1; 	// 100 milliseconds
    private double error_probability = 0.0;

    private Random rnd = new Random();

    private PacketQueue tx_queue;
    private TransportReceiver receiver;

    public static final int DEFAULT_TX_QUEUE_SIZE = 100;
    public static final int MTU = 1500;

    public BasicLink() {
	tx_queue = new PacketQueue(DEFAULT_TX_QUEUE_SIZE);
    }

    public BasicLink(int tx_qsize) {
	tx_queue = new PacketQueue(tx_qsize);
    }

    public void set_error_probability(double p) {
	if (p >= 0 && p <= 1)
	    error_probability = p;
    }

    public void set_rate(double r) {
	if (r > 0)
	    rate = r;
    }

    public void set_delay(double d) {
	if (d > 0)
	    delay = d;
    }

    void connect(TransportReceiver r) {
	receiver = r;
    }

    public void send(Packet p) {
	if (tx_queue.empty())
	    Simulation.schedule_event(p.size()/rate, new PacketDeparture(this));
	tx_queue.enqueue(p);
    }

    void process_departure() {
	Packet p = tx_queue.dequeue();
	if (error_probability == 0.0 || rnd.nextDouble() > error_probability)
	    Simulation.schedule_event(delay, new PacketDelivery(this, p));
	else 
	    System.out.printf("%f BasicLink: packet dropped %s\n",
			      Simulation.time(), p.toString());
	    
	if (!tx_queue.empty())
	    Simulation.schedule_event(tx_queue.first().size()/rate, new PacketDeparture(this));
    }

    void process_delivery(Packet p) {
	receiver.deliver(p);
    }

    public int mtu() {
	return MTU;
    }
}
