import java.util.Random;

public class PoissonSender implements Event, ApplicationSender {
    double lambda;
    TransportSender transport;
    int count;
    int max_count;
    Random rnd;
    boolean blocked;
    byte [] pending_data;
    byte [] buf;

    public int DATA_SIZE = 1400;

    public PoissonSender(int c, double l) {
	count = 0;
	max_count = c;
	lambda = l;
	rnd = new Random();
	blocked = false;
	pending_data = null;
	buf = new byte[DATA_SIZE];
	java.util.Arrays.fill(buf, (byte)0x20);
    }

    public void start(double time) {
	Simulation.schedule_event(time, this);
    }	

    public void connect(TransportSender t) {
	transport = t;
    }

    private double exp_random(double lambda) {
	return  Math.log(1 - rnd.nextDouble())/(-lambda);
    }

    private byte [] new_packet(int seq_number) {
	byte [] b = ("seq:" + seq_number).getBytes();
	for (int i = 0; i < b.length; ++i)
	    buf[i] = b[i];
	return buf;
    }

    public void stop_sender() {
	if (!blocked) {
	    blocked = true;
	    System.out.printf("%f ", Simulation.time());
	    System.out.println("Sender: blocking");
	}
    }

    public void continue_sender() {
	if (blocked) {
	    blocked = false;
	    System.out.printf("%f ", Simulation.time());
	    System.out.println("Sender: continuing");
	    if (pending_data != null)
		process();
	    else
		Simulation.schedule_event(exp_random(lambda), this);
	}
    }

    public void process() {	// action for "wake up" event
	if (count == 0) {
	    transport.open();
	    count = 1;
	    Simulation.schedule_event(exp_random(lambda), this);
	} else if (count <= max_count) {
	    if (blocked)
		return;
	    byte [] data;
	    if (pending_data != null)
		data = pending_data;
	    else
		data = new_packet(count);

	    System.out.printf("%f ", Simulation.time());
	    String display = (data.length > 8) ? (new String(data,0,8) + "...") : new String(data);
	    System.out.println("Sender:" + count + ": data=" + display + " size=" + DATA_SIZE);
	    if (!transport.send(data, DATA_SIZE)) {
		pending_data = data;
		System.out.printf("%f ", Simulation.time());
		System.out.println("Sender:" + count + ": rejected");
		return;
	    } 
	    Simulation.schedule_event(exp_random(lambda), this);
	    count += 1;
	    pending_data = null;
	} else {
	    transport.close();
	    System.out.printf("%f ", Simulation.time());
	    System.out.println("Sender: terminated.");
	}
    }
}
