//
// A very basic simulation kernel.  The Simulation class implements a
// *static* simulation engine with a static interface.
//
import java.util.ArrayList;
import java.util.PriorityQueue;

interface Event {
    public void process();
}

class ScheduledEvent implements Comparable<ScheduledEvent> {
    double time;
    Event event;

    ScheduledEvent(double t, Event e) {
	time = t;
	event = e;
    }

    public int compareTo(ScheduledEvent other) {
	if (time < other.time)
	    return -1;
	if (time > other.time)
	    return 1;
	return 0;
    }
}

public class Simulation {
    static private PriorityQueue<ScheduledEvent> events = new PriorityQueue<ScheduledEvent>();
    static private double now;
    static private ScheduledEvent [] cache = new ScheduledEvent [1000];
    static private int cache_size = 0;

    static private ScheduledEvent new_scheduled_event(double t, Event e) {
	if (cache_size == 0)
	    return new ScheduledEvent(t, e);
	else {
	    ScheduledEvent result = cache[--cache_size];
	    result.time = t;
	    result.event = e;
	    return result;
	}
    }

    static private void recyle_scheduled_event(ScheduledEvent e) {
	if (cache_size < cache.length) {
	    e.event = null;
	    cache[cache_size++] = e;
	}
    }

    static private Event next_event() {
	ScheduledEvent e = events.poll();
	if (e == null)
	    return null;
	now = e.time;
	return e.event;
    }

    static private boolean have_events() {
	return !events.isEmpty();
    }

    static public void clear() {
	events.clear();
	now = 0.0;
    }

    static public double time() { // clock (virtual)
	return now;
    }

    static public void schedule_event(double delta, Event event) {
	// ASSERT: delta >= 0
	events.offer(new_scheduled_event(now + delta, event));
    }

    static public void run() {
	while (have_events()) 
	    next_event().process();
    }
}
