import java.util.HashMap;

public class TCPReceiver implements TransportReceiver {
    Link link;
    int src_port;
    int dst_port;

    int seq_num;

    ApplicationReceiver app;

    int state;
    static final int CLOSED = 0;
    static final int LISTEN = 1;
    static final int SYN_RCVD = 2;
    static final int ESTABLISHED = 3;
    static final int CLOSE_WAIT = 4; // not yet implemented
    static final int LAST_ACK = 5;   // not yet implemented

    TCPReceiver(int src, int dst) {
	src_port = src;
	dst_port = dst;
	state = LISTEN;
    }

    public void connect(ApplicationReceiver a) {
	app = a;
    }

    public void connect(Link l) {
	link = l;
    }

    public void deliver(Packet packet) {
	if (packet instanceof TCPPacket) {
	    TCPPacket p = (TCPPacket)packet;
	    switch (state) {
	    case SYN_RCVD:
	    case LISTEN:
		if (p.syn) {
		    System.out.printf("%f ", Simulation.time());
		    System.out.println("TCPReceiver: received SYN packet seq_num=" + p.seq_num); 
		    p.ack = true;
		    p.syn = true;
		    p.ack_num = TCP.next_seq(p.seq_num, 1);
		    seq_num = p.ack_num;
		    // we should set our own sequence number, but we
		    // don't have a backward stream here, so we just set it to 0
		    p.seq_num = 0;
		    p.app_data = null;
		    p.app_size = 0;
		    p.src_port = src_port;
		    p.dst_port = dst_port;
		    link.send(p);
		    state = SYN_RCVD;
		    System.out.printf("%f ", Simulation.time());
		    System.out.println("TCPReceiver: sent SYN,ACK packet seq_num=" + p.seq_num); 
		    break;
		} else if (p.ack) {
		    state = ESTABLISHED;
		    break;
		}
		state = ESTABLISHED;
	    case ESTABLISHED:
		if (p.seq_num == seq_num) {
		    app.deliver(p.app_data, p.app_size);
		    seq_num = TCP.next_seq(seq_num, p.app_size);
		}
		p.ack_num = seq_num;
		p.ack = true;
		p.app_data = null;
		p.app_size = 0;
		p.src_port = src_port;
		p.dst_port = dst_port;
		System.out.printf("%f ", Simulation.time());
		System.out.println("TCPReceiver: sending ACK ack_num=" + p.ack_num); 
		link.send(p);
		break;
	    default:
		// work in progress
	    }
	}
    }
}
